#!/bin/sh
#
# generic build script

## HELPER FUNCTIONS
###################

die() { echo "$@"; exit 1; }
command_exists() { type "$1" >/dev/null 2>&1; }

## DETECT COMPILER ENVIRONMENT
##############################

base=$(cygpath $WD/../../../Repository)

if [[ "$MSYSTEM" != "MINGW32" && "$MSYSTEM" != "MINGW64" ]] ; then

  if [[ "$1" == "x86" ]] ; then
    export MSYSTEM=MINGW32
  elif [[ "$1" == "x64" ]] ; then
    export MSYSTEM=MINGW64
  else
    die "Build: No build environment specified (chose from: x86, x64; then add as argument) - running a $MSYSTEM shell."
  fi

  source /etc/profile
  shift
fi

if [[ "$1" == "x86" && "$MSYSTEM" != "MINGW32" ]]; then
  die "Build: Running a $MSYSTEM shell does not allow building $1 libraries."
elif [[ "$1" == "x64" && "$MSYSTEM" != "MINGW64" ]]; then
  die "Build: Running a $MSYSTEM shell does not allow building $1 libraries."
elif [[ "$1" == "x86" || "$1" == "x64" ]]; then
  shift
fi

## DETECT LN-UAC
################

if ! command_exists ln-uac; then
  cp -u $base/System/usr/bin/elevate.exe  /usr/bin/elevate.exe
  cp -u $base/System/usr/bin/ln-uac       /usr/bin/ln-uac
fi

if ! command_exists ln-uac; then
  die "Build: ln-uac not found. (tool to use elevation when creating symbolic links)"
fi

## DETECT COMPILER
##################

if [[ "$1" == "gcc" || "$1" == "clang" ]]; then
  COMPILER=$1
  shift
else
  die "Build: No compiler specified (chose from: gcc, clang; then add as argument)."
fi

if ! command_exists $COMPILER; then

  if   [[ "$MSYSTEM" == "MINGW32" && "$COMPILER" == "gcc" ]]; then
    PACKAGE=mingw-w64-i686-gcc
  elif [[ "$MSYSTEM" == "MINGW64" && "$COMPILER" == "gcc" ]]; then
    PACKAGE=mingw-w64-x86_64-gcc
  elif [[ "$MSYSTEM" == "MINGW32" && "$COMPILER" == "clang" ]]; then
    PACKAGE=mingw-w64-i686-clang
  elif [[ "$MSYSTEM" == "MINGW64" && "$COMPILER" == "clang" ]]; then
    PACKAGE=mingw-w64-x86_64-clang
  else
    die "Build: Compiler not found. Cannot determine package."
  fi

  pacman --sync --needed --noconfirm $PACKAGE || die "Build: Compiler not found. Cannot install package."
fi

if ! command_exists $COMPILER; then
  die "Build: Compiler not found."
fi

## DETECT MAKE
##############

if ! command_exists make; then
  pacman --sync --needed --noconfirm make || die "Build: Make not found. Cannot install package."
fi

if ! command_exists make; then
  die "Build: Make not found."
fi

## DETECT EXTRA PACKAGES
########################

build_packages="yasm $build_packages"

for p in $build_packages; do
  if ! command_exists $p; then
    pacman --sync --needed --noconfirm $p || die "Build: $p not found. Cannot install package."
  fi
done

## SETUP ENVIRONMENT
####################

if [[ "$build_name" = "" ]] ; then
  die "Build: Nothing to do. Use the actual build script of a library or drone. Or use build_all"
fi

if [[ "$MSYSTEM" = "MINGW32" ]] ; then
  use_host=i686-pc-mingw32;
else
  use_host=x86_64-pc-mingw64;
fi

src=$base/$build_type/$build_name/$build_version
lib=$base/Build/Win${MSYSTEM#MINGW}-$COMPILER
dst=$lib/.intermediate/$build_name

## CREATE BUILD ENVIRONMENT
###########################

mkdir -p $lib/include
mkdir -p $lib/include/$files_src_include_subname
mkdir -p $lib/lib
mkdir -p $lib/bin
mkdir -p $dst

if [[ "$build_type" = "Drones" ]]; then
  mkdir -p $dst/.extern
  test  -h $dst/.extern/include || ln-uac -s $lib/include $dst/.extern/include
  test  -h $dst/.extern/lib     || ln-uac -s $lib/lib     $dst/.extern/lib
fi

cd $dst || die "Build: Cannot change into $dst"

## COMMAND: CLEANUP
###################

if [[ "$1" == "--cleanup" ]]; then
  find $dst -type f -delete
  find $dst -type d -delete
  exit 1;
fi

## DETECT: COLD-BOOT
####################

if [[ ! -f $src/configure && "$cmd_configure" = "" ]]; then

  if [ ! -f $src/autogen.sh ]; then
    die "Build: configure not found. There is also no autogen.sh to perform a coldboot."
  fi

  # request autoconf, automake
  pacman --sync --needed --noconfirm autoconf automake libtool || die "Build: automake/autoconf not found. Cannot install package."

  # cold boot
  cd $src || die "Build: Cannot change into $src"
  ./autogen.sh

  cd $dst || die "Build: Cannot change into $dst"

fi

## CONFIGURE
############

if [[ ! -f $dst/.configured || "$1" = "--configure" ]]; then

  rm -f $dst/.configured

  if [[ "$cmd_configure" = "" ]]; then
    cmd_configure=$src/configure
  fi

  eval CC=$COMPILER cc=$COMPILER CXX=$COMPILER host=$use_host $cmd_configure $options || die "Build: configure failed";
  touch $dst/.configured

fi

## MAKE
#######

make || die "Build: make failed"

## TRANSFER
###########

for f in $files_src_include; do
  cp -u $src/$f  $lib/include
done

for f in $files_src_include_sub; do
  cp -u $src/$f  $lib/include/$files_src_include_subname
done

for f in $files_dst_include; do
  cp -u $dst/$f  $lib/include
done

for f in $files_dst_lib; do
  cp -u $dst/$f  $lib/lib
done

for f in $files_dst_bin; do
  cp -u $dst/$f  $lib/bin
done

#
# end of script