/*
 * This file is part of FFmpeg.
 *
 * FFmpeg is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * FFmpeg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with FFmpeg; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

#pragma GCC diagnostic ignored "-Wmultichar"

#include "config.h"
#include "ffdrone_win.h"

#include "libavutil/avutil.h"
#include "libavutil/internal.h"
#include "libavutil/opt.h"
#include "libavformat/avformat.h"
#include "libavformat/internal.h"
#include "libavcodec/avcodec.h"
#include "libavcodec/internal.h"

#define av_quit(code)                             fflush(stderr), Sleep(2000), ExitProcess(0xDEAD0000 + code)

const char program_name[] = "ffdrone";
const int program_birth_year = 2013;

void show_help_default(const char *opt, const char *arg);
void show_help_default(const char *opt, const char *arg) { }

// =======================================================================================================================================================================================================================================================
// STRUCTURES
// =======================================================================================================================================================================================================================================================

struct COMMAND_INIT
{
	int32_t signature;
	int32_t version;
	int32_t share_input;
	int32_t share_output;

	int32_t type;
	int32_t codec;
	int32_t threads;
	int32_t reserved;

	int32_t values[32];
};

struct COMMAND_TASK
{
	int64_t input;
	int64_t output;

	int32_t input_len;
	int32_t output_size;

	int32_t values[32];
};

struct RESULT_TASK
{
	int32_t ret;
	int32_t output_len;

	int32_t values[32];
};

#define TYPE_CONTAINER  0
#define TYPE_DECODER    1
#define TYPE_ENCODER    2

// =======================================================================================================================================================================================================================================================
// GLOABALS
// =======================================================================================================================================================================================================================================================

struct COMMAND_INIT init;
struct COMMAND_TASK task;
struct RESULT_TASK result;

static char *params;
static short paramlen;

static AVCodecContext *context_codec;
static AVFormatContext *context_format;
static int context_format_busy;

static AVCodec *codec;
static AVOutputFormat *format;
static int initialized;

static AVPacket packet;
static AVFrame frame;

// =======================================================================================================================================================================================================================================================
// STATIC METHOD - DRONE - RESET
// =======================================================================================================================================================================================================================================================

static void drone_reset()
{
	if (init.type == TYPE_CONTAINER)
	{
		avformat_free_context(context_format);
	}
	else
	{
		avcodec_close(context_codec);
		av_free(context_codec);
		free(params);
	}

	params = NULL, paramlen = 0;
	context_codec = NULL, context_format = NULL, context_format_busy = 0;
	codec = NULL, format = NULL, initialized = 0;

	memset(&packet, 0, sizeof(packet));
  memset(&frame, 0, sizeof(frame));

  memset(&init, 0, sizeof(init));
  memset(&task, 0, sizeof(task));
  memset(&result, 0, sizeof(result));
}

// =======================================================================================================================================================================================================================================================
// STATIC METHOD - DRONE - PIPE - READ
// =======================================================================================================================================================================================================================================================

static void drone_pipe_read(void *data, int length)
{
	static HANDLE Pipe; int have, pos;

	if (!Pipe)
		Pipe = GetStdHandle(STD_INPUT_HANDLE);

	for (pos = 0; pos < length; ++pos)
	{
		if (!ReadFile(Pipe, (byte *) data + pos, length - pos, (DWORD *) &have, NULL) || have == 0)
			av_quit(11);

		pos += have;
	}
}

// =======================================================================================================================================================================================================================================================
// STATIC METHOD - DRONE - PIPE - WRITE
// =======================================================================================================================================================================================================================================================

static void drone_pipe_write(const void *data, int length)
{
	static HANDLE Pipe; int have;

	if (!Pipe)
		Pipe = GetStdHandle(STD_OUTPUT_HANDLE);

	for (int Try = 0; Try < 20 * 10; ++Try)
	{
		if (WriteFile(Pipe, data, length, (DWORD *) &have, NULL))
			return;

		Sleep(100);
	}

	if (!WriteFile(Pipe, data, length, (DWORD *) &have, NULL))
		av_quit(12);
}

// =======================================================================================================================================================================================================================================================
// STATIC METHOD - DRONE - TRANSFER - INPUT
// =======================================================================================================================================================================================================================================================

static byte VIDEO_SURFACE[4096 * 2048 * 8];

static void drone_transfer_input()
{
	if (!init.share_input && task.input_len <= LONG_MAX) // using pipe
	{
		static byte *memory_input;

		if (!memory_input)
			memory_input = av_malloc(1024 << 10);

		task.input = (int64_t) memory_input;

		drone_pipe_read(memory_input, task.input_len);
	}

	if (!init.share_output)
	{
		static byte *memory_output;

		if (!memory_output)
			memory_output = av_malloc(1024 << 10);

		task.output = (int64_t) memory_output;
	}

	// transfer into shared memory (video)
	if (init.share_input && init.type == TYPE_ENCODER && codec->type == AVMEDIA_TYPE_VIDEO)
	{
		enum AVPixelFormat format = task.values[2] & 8 ? AV_PIX_FMT_YUV420P10 : AV_PIX_FMT_YUV420P;

		int isSampling10 = codec->type == AVMEDIA_TYPE_VIDEO && format == AV_PIX_FMT_YUV420P10;

		if (isSampling10)
		{
			int p, d, r, w, height = init.values[1], plane = 0, plane_sec = 0, pitch = task.values[1]; unsigned short* frame_data[3];

			byte *input = (byte *) task.input;
			byte *output = (byte *) VIDEO_SURFACE;

      if (height > 2160)
        height = 2160;
      if (pitch > 4096 * 2)
        pitch = 4096 * 2;

      pitch /= 2; // in shorts
			plane = pitch * height / 4;

			frame_data[0] = (unsigned short *) output;
			frame_data[1] = (unsigned short *) output + plane * 4;
			frame_data[2] = (unsigned short *) output + plane * 5;

			plane = pitch * height, plane_sec = plane / 4;

			if (codec->type == AVMEDIA_TYPE_VIDEO)
			{
				unsigned short *Y = (unsigned short *) input;
				unsigned short *U = (unsigned short *) input + plane;
				unsigned short *V = U + 1;

				for (p = 0; p < 3; ++p, plane = plane_sec) if (frame_data[p])
				{
					unsigned short *inp = p == 0 ? Y : p == 1 ? U : V; d = p == 0 ? 1 : 2;

					for (r = 0, w = 0; r < plane; ++r, w += d)
						frame_data[p][r] = (inp[w] + (1 << 5)) >> 6; // /*p == 0 ? 64 : p == 1 ? 255 : 64; */ (inp[w] + (1 << 5)) >> 6;
				}
			}

			//task.values[1] /= 2;
			task.input = output;
		}
		else
		{
			int p, d, r, w, height = init.values[1], plane = 0, plane_sec = 0, pitch = task.values[1]; byte* frame_data[3];

			byte *input = (byte *) task.input;
			byte *output = (byte *) VIDEO_SURFACE;

      if (height > 2160)
        height = 2160;
      if (pitch > 4096)
        pitch = 4096;

		  plane = pitch * height / 4;

			frame_data[0] = output;
			frame_data[1] = output + plane * 4;
			frame_data[2] = output + plane * 5;

			plane = pitch * height, plane_sec = plane / 4;

			if (codec->type == AVMEDIA_TYPE_VIDEO)
			{
				unsigned char *Y = (unsigned char *) input;
				unsigned char *U = (unsigned char *) (input + plane);
				unsigned char *V = U + 1;

				//char S[255];
				//sprintf(S, "W: %d, H: %d, P: %d, plane: %d, plane_sec: %d\n", init.values[0], init.values[1], task.values[1], plane, plane_sec);

				//OutputDebugString(S);
				//memset(VIDEO_SURFACE + 1920 * 1080 * 0, 0xCC, 1920 * 1080);
				//memset(VIDEO_SURFACE + 1920 * 1080 * 1, 0xCC, 1920 * 1080);

				for (p = 0; p < 3; ++p, plane = plane_sec) if (frame_data[p])
				{
					unsigned char *inp = p == 0 ? Y : p == 1 ? U : V; d = p == 0 ? 1 : 2;

					for (r = 0, w = 0; r < plane; ++r, w += d)
						frame_data[p][r] = inp[w];
				}
			}

			task.input = output;
		}
	}
}

// =======================================================================================================================================================================================================================================================
// STATIC METHOD - DRONE - TRANSFER - OUTPUT
// =======================================================================================================================================================================================================================================================

static void drone_transfer_output()
{
	byte *output = (byte *) task.output;

	// transfer into shared memory (video)
	if (init.share_output && init.type == TYPE_DECODER && result.output_len)
	{
		int p, d, r, w, plane = 0, plane_sec = 0, isSampling10, isOutput16;

		if (codec->type == AVMEDIA_TYPE_VIDEO && frame.linesize[0])
			plane = frame.linesize[0] * frame.height, plane_sec = plane / 4;
		if (codec->type == AVMEDIA_TYPE_AUDIO && frame.linesize[0])
			plane = frame.linesize[0], plane_sec = plane;

		isSampling10 = codec->type == AVMEDIA_TYPE_VIDEO && frame.format == AV_PIX_FMT_YUV420P10LE;
		isOutput16 = (task.values[2] & 2) != 0;

		if (isSampling10 && isOutput16)
		{
			unsigned short *Y = (unsigned short *) output;
			unsigned short *U = (unsigned short *) (output + plane);
			unsigned short *V = U + 1;

			result.values[0] /= 1, plane /= 2, plane_sec /= 2;

			for (p = 0; p < 3; ++p, plane = plane_sec) if (frame.data[p])
			{ 
				unsigned short *out = p == 0 ? Y : p == 1 ? U : V; d = p == 0 ? 1 : 2;

				for (r = 0, w = 0; r < plane; ++r, w += d)
					out[w] = /* (p + 1) << 8 | (r & 0xFF); */ (*(unsigned short *) &frame.data[p][2 * r]) << 6; // (10-bit to 16-bit)
			}
		}
		else if (isSampling10)
		{
			unsigned char *Y = (unsigned char *) output;
			unsigned char *U = (unsigned char *) (output + plane / 2);
			unsigned char *V = U + 1;

			result.values[0] /= 2, plane /= 2, plane_sec /= 2;

			for (p = 0; p < 3; ++p, plane = plane_sec) if (frame.data[p])
			{
				unsigned char *out = p == 0 ? Y : p == 1 ? U : V; d = p == 0 ? 1 : 2;

				for (r = 0, w = 0; r < plane; ++r, w += d)
					out[w] = /* (p + 1) << 8 | (r & 0xFF); */ (*(unsigned short *) &frame.data[p][2 * r] + 2) >> 2; // (10-bit to 8-bit)
			}
		}
		else if (codec->type == AVMEDIA_TYPE_VIDEO)
		{
			unsigned char *Y = (unsigned char *) output;
			unsigned char *U = (unsigned char *) (output + plane);
			unsigned char *V = U + 1;

			for (p = 0; p < 3; ++p, plane = plane_sec) if (frame.data[p])
			{
				unsigned char *out = p == 0 ? Y : p == 1 ? U : V; d = p == 0 ? 1 : 2;

				for (r = 0, w = 0; r < plane; ++r, w += d)
					out[w] = /* (p + 1) << 8 | (r & 0xFF); */ (*(unsigned char *) &frame.data[p][r]);
			}
		}
		else // other
		{
			for (p = 0; p < 6; ++p, plane = plane_sec) if (frame.data[p])
			{
				memcpy(output, frame.data[p], plane);

				output += plane;
			}
		}
	}

	drone_pipe_write(&result, sizeof(struct RESULT_TASK));
}

// =======================================================================================================================================================================================================================================================
// STATIC METHOD - DRONE - CODEC - INIT - DECODE - VIDEO
// =======================================================================================================================================================================================================================================================

static void drone_codec_decode_video()
{
	int got = 0, pitch, plane;

	packet.data = task.input_len ? (byte *) task.input : NULL; // special case for decoder flush
	packet.size = task.input_len;

	result.ret = avcodec_decode_video2(context_codec, &frame, &got, &packet);

	if (got && result.ret != packet.size && (result.ret & 0x40000000) == 0)
		result.ret = AVERROR_INVALIDDATA;

	if (got && !frame.key_frame && (task.values[2] & 4) != 0 && context_codec->codec_id == AV_CODEC_ID_HEVC)
		got = 0;

	else if (!got && (task.values[2] & 4) != 0)
	{
		packet.data = NULL;
		packet.size = 0;

		result.ret = avcodec_decode_video2(context_codec, &frame, &got, &packet);
		if (result.ret == 0)
			result.ret = task.input_len;
	}

	if (!got && result.ret < 0 && (task.values[2] & 4) != 0)
		result.ret = 0;

	pitch = got ? frame.linesize[0] : 0;
	plane = pitch * context_codec->height / 4;
	
	if (!got)
	{
		byte *output = (byte *) task.output;

		int plane = context_codec->width * context_codec->height;

		memset(output,         0x00, plane / 1);
		memset(output + plane, 0x80, plane / 2);
	}

	result.output_len = plane * 6; // YUV 420

	result.values[0] = pitch;
	result.values[1] = frame.key_frame;
	result.values[2] = frame.coded_picture_number;
	result.values[3] = frame.display_picture_number;
	result.values[4] = context_codec->frame_number;
	result.values[5] = frame.repeat_pict;
	result.values[6] = frame.interlaced_frame;
	result.values[7] = frame.top_field_first;
	result.values[8] = frame.flags;
}

// =======================================================================================================================================================================================================================================================
// STATIC METHOD - DRONE - CODEC - DECODE - AUDIO
// =======================================================================================================================================================================================================================================================

static void drone_codec_decode_audio()
{
	int got = 0, pitch;

	packet.data = task.input_len ? (byte *) task.input : NULL; // special case for decoder flush
	packet.size = task.input_len;

	result.ret = avcodec_decode_audio4(context_codec, &frame, &got, &packet);

	if (got && result.ret != packet.size)
		result.ret = AVERROR_INVALIDDATA;

	pitch = got ? frame.linesize[0] : 0;

	result.output_len = pitch;
}

// =======================================================================================================================================================================================================================================================
// STATIC METHOD - DRONE - CODEC - ENCODE - VIDEO
// =======================================================================================================================================================================================================================================================

static void drone_codec_encode_video()
{
	int got = 0, pitch = task.values[1], plane; byte *input;

	plane = pitch * context_codec->height / 4;
	input = (byte *) task.input;

	frame.width = init.values[0];
	frame.height = init.values[1];
	frame.format = task.values[2] & 8 ? AV_PIX_FMT_YUV420P10 : AV_PIX_FMT_YUV420P;

	frame.data[0] = input;
	frame.data[1] = input + plane * 4;
	frame.data[2] = input + plane * 5;

	frame.linesize[0] = pitch;
	frame.linesize[1] = pitch / 2;
	frame.linesize[2] = pitch / 2;

	frame.pts = task.values[0]; // libx264 requires pts

	packet.data = (byte *) task.output;
	packet.size = task.output_size;

	if (context_codec->codec_id == AV_CODEC_ID_H264 && (task.values[2] & 1) != 0)
		ff_X264_intra_refresh(context_codec);

	result.ret = avcodec_encode_video2(context_codec, &packet, task.input_len ? &frame : NULL, &got);
	result.output_len = packet.size;

	result.values[0] = (int32_t) (size_t) context_codec->extradata;
	result.values[1] = context_codec->extradata_size;
#if FF_API_CODED_FRAME
	FF_DISABLE_DEPRECATION_WARNINGS
		result.values[2] = context_codec->coded_frame && context_codec->coded_frame->pict_type;
	FF_ENABLE_DEPRECATION_WARNINGS
#endif
		result.values[3] = (packet.flags & AV_PKT_FLAG_KEY) != 0;
	result.values[4] = packet.pts;
	result.values[5] = packet.dts;

  if (context_codec->codec_id == AV_CODEC_ID_HEVC)
    result.values[4]++; // one more

	result.values[31] = (int32_t) ((size_t) context_codec->extradata >> 32);
}

// =======================================================================================================================================================================================================================================================
// STATIC METHOD - DRONE - CODEC - ENCODE - SUBTITLE
// =======================================================================================================================================================================================================================================================

static void drone_codec_encode_subtitle() // time, duration, left, top, width, height,
{
	AVSubtitle sub;
	AVSubtitleRect rect;
	AVSubtitleRect* rects[1];

	sub.format = 0;
	sub.start_display_time = 0;
	sub.end_display_time = task.values[1];
	sub.num_rects = 1;
	sub.rects = rects;
	sub.pts = task.values[0];

	rects[0] = &rect;

	rect.x = task.values[2];
	rect.y = task.values[3];
	rect.w = task.values[4];
	rect.h = task.values[5];
	rect.nb_colors = 256;
	rect.type = SUBTITLE_BITMAP;
	rect.text = NULL;
	rect.ass = NULL;
	rect.flags = 0;

	rect.pict.data[0] = (uint8_t *) task.input;
	rect.pict.data[1] = (uint8_t *) task.input + rect.w * rect.h;
	rect.pict.linesize[0] = rect.w;
	rect.pict.linesize[1] = 4;

	result.ret = avcodec_encode_subtitle(context_codec, (uint8_t *) task.output, task.output_size, &sub);

	if (result.ret > 0)
		result.output_len = result.ret, result.ret = 0;
	else
		result.output_len = 0;
}

// =======================================================================================================================================================================================================================================================
// STATIC METHOD - DRONE - CODEC - ENCODE - AUDIO
// =======================================================================================================================================================================================================================================================

static byte input_audio[8192 * 8 * 4];
static int audio_downmix;

static void drone_codec_encode_audio()
{
	int got = 0, samples = task.values[1], plane;

	if (context_codec->codec->sample_fmts[0] == AV_SAMPLE_FMT_S32)
	{
		for (int nn = 0, pp = 0; nn < samples; nn += audio_downmix) for (int cc = 0; cc < context_codec->channels; ++cc)
		{
			float f = ((float *) task.input)[cc * samples + nn];

			if (audio_downmix == 2)
			{
				float f1 = ((float *) task.input)[cc * samples + nn + 0];
				float f2 = ((float *) task.input)[cc * samples + nn + 1];

				f = (f1 + f2) / 2;
			}

			if (audio_downmix == 4)
			{
				float f1 = ((float *) task.input)[cc * samples + nn + 0];
				float f2 = ((float *) task.input)[cc * samples + nn + 1];
				float f3 = ((float *) task.input)[cc * samples + nn + 2];
				float f4 = ((float *) task.input)[cc * samples + nn + 3];

				f = (f1 + f2 + f3 + f4) / 4;
			}

			int64_t s = (int64_t) (f * INT_MAX);
			if (s > INT_MAX) s = INT_MAX;
			if (s < INT_MIN) s = INT_MIN;

			((int *) input_audio)[pp++] = s;
		}

		plane = samples / audio_downmix * context_codec->channels * sizeof(int);

		for (int nn = 0; nn < 1; ++nn)
		{
			frame.data[nn] = (byte *) input_audio;
			frame.linesize[nn] = plane;
		}
	}
	else if (context_codec->codec->sample_fmts[0] == AV_SAMPLE_FMT_S16)
	{
		for (int nn = 0, pp = 0; nn < samples; nn += audio_downmix) for (int cc = 0; cc < context_codec->channels; ++cc)
		{
			float f = ((float *) task.input)[cc * samples + nn];

			if (audio_downmix == 2)
			{
				float f1 = ((float *) task.input)[cc * samples + nn + 0];
				float f2 = ((float *) task.input)[cc * samples + nn + 1];

				f = (f1 + f2) / 2;
			}

			if (audio_downmix == 4)
			{
				float f1 = ((float *) task.input)[cc * samples + nn + 0];
				float f2 = ((float *) task.input)[cc * samples + nn + 1];
				float f3 = ((float *) task.input)[cc * samples + nn + 2];
				float f4 = ((float *) task.input)[cc * samples + nn + 3];

				f = (f1 + f2 + f3 + f4) / 4;
			}

			int s = (int) (f * MAXSHORT);
			if (s > MAXSHORT) s = MAXSHORT;
			if (s < MINSHORT) s = MINSHORT;

			((short *) input_audio)[pp++] = s;
		}

		plane = samples / audio_downmix * context_codec->channels * sizeof(short);

		for (int nn = 0; nn < 1; ++nn)
		{
			frame.data[nn] = (byte *) input_audio;
			frame.linesize[nn] = plane;
		}
	}
	else if (context_codec->codec->sample_fmts[0] == AV_SAMPLE_FMT_FLTP)
	{
		for (int cc = 0, pp = 0; cc < context_codec->channels; ++cc) for (int nn = 0; nn < samples; nn += audio_downmix) // inverted channel/sample
		{
			float f = ((float *) task.input)[cc * samples + nn];

			if (audio_downmix == 2)
			{
				float f1 = ((float *) task.input)[cc * samples + nn + 0];
				float f2 = ((float *) task.input)[cc * samples + nn + 1];

				f = (f1 + f2) / 2;
			}

			if (audio_downmix == 4)
			{
				float f1 = ((float *) task.input)[cc * samples + nn + 0];
				float f2 = ((float *) task.input)[cc * samples + nn + 1];
				float f3 = ((float *) task.input)[cc * samples + nn + 2];
				float f4 = ((float *) task.input)[cc * samples + nn + 3];

				f = (f1 + f2 + f3 + f4) / 4;
			}

			((float *) input_audio)[pp++] = f;
		}

		plane = samples / audio_downmix * sizeof(float);

		for (int nn = 0; nn < context_codec->channels; ++nn)
		{
			frame.data[nn] = (byte *) input_audio + nn * plane;
			frame.linesize[nn] = plane;
		}
	}
	else if (context_codec->codec->sample_fmts[0] == AV_SAMPLE_FMT_FLTP)
	{
		plane = samples * sizeof(float);

		for (int nn = 0; nn < context_codec->channels; ++nn)
		{
			frame.data[nn] = (byte *) task.input + nn * plane;
			frame.linesize[nn] = plane;
		}
	}
	else if (context_codec->codec->sample_fmts[0] == AV_SAMPLE_FMT_DBL)
	{
		plane = samples * sizeof(float);

		for (int nn = 0; nn < context_codec->channels; ++nn)
		{
			frame.data[nn] = (byte *) task.input + nn * plane;
			frame.linesize[nn] = plane;
		}
	}

	frame.nb_samples = samples;

	packet.data = (byte *) task.output;
	packet.size = task.output_size;

	result.ret = avcodec_encode_audio2(context_codec, &packet, &frame, &got);
	result.output_len = packet.size;

	// manually set
	packet.pts = context_codec->frame_number;
	packet.dts = 0xFFFFFFFF;

	result.values[0] = (int32_t) (size_t) context_codec->extradata;
	result.values[1] = context_codec->extradata_size;
	result.values[2] = AV_PICTURE_TYPE_NONE;
	result.values[3] = (packet.flags & AV_PKT_FLAG_KEY) != 0;
	result.values[4] = packet.pts;
	result.values[5] = packet.dts;
}

// =======================================================================================================================================================================================================================================================
// STATIC METHOD - DRONE - CODEC - DECODE - VIDEO - INIT
// =======================================================================================================================================================================================================================================================

static void drone_codec_decode_video_init(AVDictionary **dict) // VALUES: width, height
{
	// decoder VC-1 ==> a real pain in the ass
	if (context_codec->codec_id == AV_CODEC_ID_VC1)
	{
		context_codec->width = init.values[0];
		context_codec->height = init.values[1];
		context_codec->extradata = (byte *) task.input;
		context_codec->extradata_size = task.input_len;
	}

	if (init.values[2] & 1)
	{
		context_codec->flags |= CODEC_FLAG_USE_DTS_ORDER;
	}

	// decoder H.264 ==> needs some speedup
	if (context_codec->codec_id == AV_CODEC_ID_H264)
	{
		context_codec->flags2 |= CODEC_FLAG2_FAST; // Cannot parallelize deblocking type 1, decoding such frames in sequential order -> switching
	}
}

// =======================================================================================================================================================================================================================================================
// STATIC METHOD - DRONE - CODEC - DECODE - AUDIO - INIT
// =======================================================================================================================================================================================================================================================

static void drone_codec_decode_audio_init(AVDictionary **dict) // VALUES: channels
{
	int request_channels = init.values[0];

	context_codec->request_channel_layout =
		request_channels == 6 ? AV_CH_LAYOUT_5POINT1 :
		request_channels == 2 ? AV_CH_LAYOUT_STEREO :
		request_channels == 1 ? AV_CH_LAYOUT_MONO : AV_CH_LAYOUT_STEREO;
}

// =======================================================================================================================================================================================================================================================
// STATIC METHOD - DRONE - CODEC - ENCODE - VIDEO - INIT
// =======================================================================================================================================================================================================================================================

static void drone_codec_encode_video_init(AVDictionary **dict) // VALUES: width, height, aspect-x, aspect-y, rate-num, rate-den, bitrate
{
	context_codec->width = init.values[0];
	context_codec->height = init.values[1];
	context_codec->pix_fmt = init.values[7] & 8 ? AV_PIX_FMT_YUV420P10 : AV_PIX_FMT_YUV420P;
	context_codec->sample_aspect_ratio.num = init.values[2];
	context_codec->sample_aspect_ratio.den = init.values[3];
	context_codec->time_base.num = init.values[5]; // time-base, not fps
	context_codec->time_base.den = init.values[4];
	context_codec->bit_rate = init.values[6];
	context_codec->flags |= CODEC_FLAG_QSCALE;

	context_codec->coder_type = FF_CODER_TYPE_AC;
	context_codec->gop_size = 24;
	context_codec->max_b_frames = 2;

	if (context_codec->codec_id == AV_CODEC_ID_H264)
	{
		av_opt_set(context_codec->priv_data, "profile", "high", 0);
		av_opt_set(context_codec->priv_data, "level", "41", 0);
	}

	if (context_codec->codec_id == AV_CODEC_ID_HEVC)
	{
	}

	av_opt_set(context_codec->priv_data, "preset", "ultrafast", 0);

	if (init.values[7] & 1)
		context_codec->flags |= CODEC_FLAG_GLOBAL_HEADER;

	context_codec->sample_aspect_ratio = av_mul_q(context_codec->sample_aspect_ratio, (AVRational) { context_codec->height, context_codec->width }); // revert sar->dar

	// DICTIONARY
	// ==========

	if (init.values[7] & 0x1000)
	{
		char *parameters = params, *name, *value; int len = paramlen, pos;

		for (pos = 0; pos < len; )
		{
			int a, b, flags;

			a = strlen(parameters + pos);
			if (pos + a + 1 > len)
				break;

			b = strlen(parameters + pos + a + 1);
			if (pos + a + 1 + b + 1 + 4 > len)
				break;

			name = (char *) parameters + pos;
			value = (char *) parameters + pos + a + 1;
			flags = *(int *) (parameters + pos + a + 1 + b + 1);

			if (strcmp(name, "x264opts") == 0) {
				av_opt_set(context_codec->priv_data, name, value, flags);
			}

			if (strcmp(name, "preset") == 0) {
				av_opt_set(context_codec->priv_data, name, value, flags);
			}

			av_opt_set(dict, name, value, flags);
			av_opt_set(context_codec->priv_data, name, value, flags);

			pos += a + 1 + b + 1 + 4;
		}
	}
}

// =======================================================================================================================================================================================================================================================
// STATIC METHOD - DRONE - CODEC - ENCODE - AUDIO - INIT
// =======================================================================================================================================================================================================================================================

static void drone_codec_encode_audio_init(AVDictionary **dict) // VALUES: channels, sample-rate, bitrate
{
	context_codec->sample_fmt =
		context_codec->codec_id == AV_CODEC_ID_DTS ? AV_SAMPLE_FMT_S32 :
		context_codec->codec_id == AV_CODEC_ID_AC3 ? AV_SAMPLE_FMT_FLTP :
		context_codec->codec_id == AV_CODEC_ID_AAC ? AV_SAMPLE_FMT_S16 :
		AV_SAMPLE_FMT_FLTP;
	context_codec->channels = init.values[0] == 8 ? 6 : init.values[0];
	context_codec->channel_layout = context_codec->channels == 8 ? AV_CH_LAYOUT_7POINT1 : context_codec->channels == 6 ? AV_CH_LAYOUT_5POINT1 : context_codec->channels == 2 ? AV_CH_LAYOUT_STEREO : context_codec->channels == 1 ? AV_CH_LAYOUT_MONO : AV_CH_LAYOUT_STEREO;
	context_codec->sample_rate = init.values[1];
	context_codec->bits_per_coded_sample = init.values[4];
	context_codec->bit_rate = init.values[2];

	audio_downmix = context_codec->sample_rate / 48000;
	context_codec->sample_rate /= audio_downmix;

	if (context_codec->codec_id == AV_CODEC_ID_AAC && context_codec->channels == 6)
		context_codec->channel_layout = AV_CH_LAYOUT_5POINT1_BACK;

	if (context_codec->codec_id == AV_CODEC_ID_AAC)
	{
		av_opt_set(context_codec->priv_data, "profile", "HE-AAC", 0);
	}

	if (init.values[3] & 1)
		context_codec->flags |= CODEC_FLAG_GLOBAL_HEADER;

	context_codec->strict_std_compliance = FF_COMPLIANCE_EXPERIMENTAL; // we are heroes
}

// =======================================================================================================================================================================================================================================================
// STATIC METHOD - DRONE - CODEC - ENCODE - SUBTITLE - INIT
// =======================================================================================================================================================================================================================================================

static void drone_codec_encode_subtitle_init(AVDictionary **dict) // VALUES: none
{
}

// =======================================================================================================================================================================================================================================================
// STATIC METHOD - DRONE - CODEC - INIT
// =======================================================================================================================================================================================================================================================

static void drone_codec_init()
{
	AVDictionary *dict = NULL;

	//if (init.type == TYPE_ENCODER && codec->type == AVMEDIA_TYPE_VIDEO && init.threads > 12)
	//	init.threads = 12;

	context_codec->debug |= FF_DEBUG_PICT_INFO;
	context_codec->thread_count = init.threads;

	if (init.type == TYPE_DECODER && codec->type == AVMEDIA_TYPE_VIDEO)
		drone_codec_decode_video_init(&dict);
	else if (init.type == TYPE_DECODER && codec->type == AVMEDIA_TYPE_AUDIO)
		drone_codec_decode_audio_init(&dict);

	else if (init.type == TYPE_ENCODER && codec->type == AVMEDIA_TYPE_VIDEO)
		drone_codec_encode_video_init(&dict);
	else if (init.type == TYPE_ENCODER && codec->type == AVMEDIA_TYPE_AUDIO)
		drone_codec_encode_audio_init(&dict);
	else if (init.type == TYPE_ENCODER && codec->type == AVMEDIA_TYPE_SUBTITLE)
		drone_codec_encode_subtitle_init(&dict);

	else
	{
		av_log(NULL, AV_LOG_ERROR, "codec: initializer not started for type=%u, codec.type=%u\n", init.type, codec->type);
		av_quit(5);
	}

	result.ret = avcodec_open2(context_codec, codec, &dict);

	if (result.ret != 0 || !context_codec->codec)
	{
		av_log(NULL, AV_LOG_ERROR, "codec: open failed for type=%u, codec.type=%u, result=%d\n", init.type, codec->type, result.ret);
		av_quit(3);
	}

	av_dict_free(&dict);
}

// =======================================================================================================================================================================================================================================================
// STATIC METHOD - DRONE - CODEC
// =======================================================================================================================================================================================================================================================

static void drone_codec()
{
	if (!initialized)
		drone_codec_init(), initialized = 1;

	// data is cloned, we must not free it
	memset(&frame, 0, sizeof(AVFrame));
	av_frame_unref(&frame);

	memset(&packet, 0, sizeof(AVPacket));
	av_init_packet(&packet);

	if (init.type == TYPE_DECODER && codec->type == AVMEDIA_TYPE_VIDEO)
		drone_codec_decode_video();
	if (init.type == TYPE_DECODER && codec->type == AVMEDIA_TYPE_AUDIO)
		drone_codec_decode_audio();

	if (init.type == TYPE_ENCODER && codec->type == AVMEDIA_TYPE_VIDEO)
		drone_codec_encode_video();
	if (init.type == TYPE_ENCODER && codec->type == AVMEDIA_TYPE_AUDIO)
		drone_codec_encode_audio();
	if (init.type == TYPE_ENCODER && codec->type == AVMEDIA_TYPE_SUBTITLE)
		drone_codec_encode_subtitle();
}

// =======================================================================================================================================================================================================================================================
// STATIC METHOD - DRONE - FORMAT - STREAM
// =======================================================================================================================================================================================================================================================

static void drone_format_stream()
{
	AVStream *st; AVCodec *codec; int id, codec_id;

	id = task.values[0];
	codec_id = task.values[1];

	switch (codec_id)
	{
		case 'h262': codec_id = AV_CODEC_ID_MPEG2VIDEO;  break;
		case 'h264': codec_id = AV_CODEC_ID_H264;        break;
		case 'h265': codec_id = AV_CODEC_ID_HEVC;        break;
		case 'vc1':  codec_id = AV_CODEC_ID_VC1;         break;

		case 'avc':  codec_id = AV_CODEC_ID_H264;        break;
		case 'hevc': codec_id = AV_CODEC_ID_HEVC;        break;

		case 'mp2':  codec_id = AV_CODEC_ID_MP2;         break;
		case 'mp3':  codec_id = AV_CODEC_ID_MP3;         break;
		case 'aac':  codec_id = AV_CODEC_ID_AAC;         break;
		case 'ac3':  codec_id = AV_CODEC_ID_AC3;         break;
		case 'ec3':  codec_id = AV_CODEC_ID_EAC3;        break;
		case 'mlp':  codec_id = AV_CODEC_ID_TRUEHD;      break;
		case 'dts':  codec_id = AV_CODEC_ID_DTS;         break;

		case 'lp16': codec_id = AV_CODEC_ID_PCM_S16BE;   break;
		case 'lp20': codec_id = AV_CODEC_ID_PCM_S24BE;   break;
		case 'lp24': codec_id = AV_CODEC_ID_PCM_S24BE;   break;

		case 'dvd':  codec_id = AV_CODEC_ID_DVD_SUBTITLE; break;
		case 'pg':   codec_id = AV_CODEC_ID_HDMV_PGS_SUBTITLE; break;
		case 'pgs':  codec_id = AV_CODEC_ID_HDMV_PGS_SUBTITLE; break;

		default:
			av_log(NULL, AV_LOG_ERROR, "codec: unknown request %u [%x]\n", codec_id, codec_id);
	}

	// id, codec_id, frame_rate_n, frame_rate_d
	codec = avcodec_find_encoder(codec_id);
	if (!codec)
		codec = avcodec_find_decoder(codec_id);

	// codec not found
	if (!codec)
	{
		AVCodec *c;

		av_log(NULL, AV_LOG_ERROR, "codec: request %u [%x]\n", codec_id, codec_id);

		for (c = av_codec_next(NULL); c; c = av_codec_next(c))
		{
			av_log(NULL, AV_LOG_ERROR, "codec: %s (%s) ; %u, %u\n", c->name, c->long_name, c->type, c->id);
		}

		av_quit(4);
	}

	//if (context_format_busy)
	//	av_quit(0x103);

	st = avformat_new_stream(context_format, codec);
	st->id = id;
	st->codec->codec_type = codec->type;
	st->codec->codec_id = codec->id;
	st->codec->time_base.den = task.values[2];
	st->codec->time_base.num = task.values[3]; // time-base, not fps

	if (!st->codec->time_base.num || !st->codec->time_base.den)
		av_quit(0x101);

	if (task.input)
	{
		st->codec->flags |= CODEC_FLAG_GLOBAL_HEADER;
		st->codec->extradata_size = task.input_len;
		st->codec->extradata = av_malloc(task.input_len);

		memcpy(st->codec->extradata, (byte *) task.input, task.input_len);
	}

	// width, height, aspect_ratio_x, aspect_ratio_y
	if (codec->type == AVMEDIA_TYPE_VIDEO)
	{
		st->codec->width = task.values[4];
		st->codec->height = task.values[5];
		st->codec->sample_aspect_ratio.num = task.values[6];
		st->codec->sample_aspect_ratio.den = task.values[7];

		if (!st->codec->sample_aspect_ratio.num)
			av_quit(0x105);
		if (!st->codec->sample_aspect_ratio.den)
			av_quit(0x106);

		// from DAR to SAR
		st->codec->sample_aspect_ratio = av_mul_q(st->codec->sample_aspect_ratio, (AVRational) { st->codec->height, st->codec->width });

		st->sample_aspect_ratio = st->codec->sample_aspect_ratio; // have to make avformat_write_header() happy
	}

	// sampling_rate, channels, frame_size
	if (codec->type == AVMEDIA_TYPE_AUDIO)
	{
		st->codec->sample_rate = task.values[4];
		st->codec->channels = task.values[5];
		st->codec->bits_per_raw_sample = task.values[7];
		st->codec->bits_per_coded_sample = task.values[7];
		st->codec->frame_size = task.values[6];

		if (!st->codec->channels)
			av_quit(0x113);
		if (!st->codec->sample_rate)
			av_quit(0x112);
		if (!st->codec->frame_size)
			av_quit(0x111);
	}

	if (task.values[8] & 0x0100)
		st->disposition |= AV_DISPOSITION_DEFAULT;
	if (task.values[8] & 0x0200)
		st->disposition |= AV_DISPOSITION_FORCED;

	// parameters
	if (task.values[8] & 0x1000)
	{
		char *parameters; int len = 0, pos;

		// read
		drone_pipe_read(&len, 2);

		parameters = (char *) malloc(len + 4);
		drone_pipe_read(parameters, len);

		*(int *) (parameters + len) = 0;

		// extract
		for (pos = 0; pos < len; )
		{
			int a, b, flags;

			a = strlen(parameters + pos);
			if (pos + a + 1 > len)
				break;

			b = strlen(parameters + pos + a + 1);
			if (pos + a + 1 + b + 1 + 4 > len)
				break;

			flags = *(int *) (parameters + pos + a + 1 + b + 1);

			av_dict_set(&st->metadata, parameters + pos, parameters + pos + a + 1, flags);

			pos += a + 1 + b + 1 + 4;
		}

		free(parameters);
	}

	result.ret = 0;
}

// =======================================================================================================================================================================================================================================================
// STATIC METHOD - DRONE - FORMAT - HEADER
// =======================================================================================================================================================================================================================================================

static void drone_format_header()
{
	char filename[128];

	// parameters
	if (task.values[1] & 0x1000)
	{
		char *parameters; int len = 0, pos;

		// read
		drone_pipe_read(&len, 2);

		parameters = (char *) malloc(len + 4);
		drone_pipe_read(parameters, len);

		*(int *) (parameters + len) = 0;

		// extract
		for (pos = 0; pos < len; )
		{
			int a, b, flags;

			a = strlen(parameters + pos);
			if (pos + a + 1 > len)
				break;

			b = strlen(parameters + pos + a + 1);
			if (pos + a + 1 + b + 1 + 4 > len)
				break;

			flags = *(int *) (parameters + pos + a + 1 + b + 1);

			if (strcmp(parameters + pos, "chapter") == 0)
			{
				int index = flags;
				int64_t start = *(int64_t *) (parameters + pos + a + 1 + b + 1 + 4);
				int64_t end = *(int64_t *) (parameters + pos + a + 1 + b + 1 + 4 + 8);
				const char *title = parameters + pos + a + 1;

				avpriv_new_chapter(context_format, index, (AVRational) { 1, 10000000 }, start, end, title);

				pos += 8 + 8;
			}
			else
			{
				//av_log(NULL, AV_LOG_INFO, "format-dict: %s, %s, %u\n", parameters + pos, parameters + pos + a + 1, flags);

				av_dict_set(&context_format->metadata, parameters + pos, parameters + pos + a + 1, flags);
			}

			pos += a + 1 + b + 1 + 4;
		}

		free(parameters);
	}

	sprintf(filename, "\\device\\handle\\%u", task.values[0]);

	result.ret = avio_open(&context_format->pb, filename, AVIO_FLAG_WRITE);
	if (result.ret != 0)
		return;

	result.ret = avformat_write_header(context_format, NULL);
}

// =======================================================================================================================================================================================================================================================
// STATIC METHOD - DRONE - FORMAT - FOOTER
// =======================================================================================================================================================================================================================================================

static void drone_format_footer()
{
	result.ret = av_write_trailer(context_format);
	if (result.ret != 0)
		return;

	avio_close(context_format->pb);
}

// =======================================================================================================================================================================================================================================================
// STATIC METHOD - DRONE - FORMAT - PACKET
// =======================================================================================================================================================================================================================================================

int marker_duration;

static void drone_format_packet()
{
	AVStream *st; int id, idx, size; byte *data;

	context_format_busy = 1;

	idx = ff_find_stream_index(context_format, task.values[0]);
	if (idx == -1)
		av_quit(0x121);

	id = task.values[0];
	st = context_format->streams[idx];

	if (!st)
		av_quit(0x122);

	size = task.input_len;
	data = av_malloc(size);
	if (!data)
		av_quit(0x1A0);

	memcpy(data, (byte *) task.input, size);

	av_init_packet(&packet);
	packet.buf = av_buffer_create(data, size, av_buffer_default_free, NULL, 0);
	packet.data = packet.buf->data;
	packet.size = packet.buf->size;

	packet.stream_index = st->index;
	packet.pts = task.values[1];
	packet.dts = task.values[2];
	packet.duration = task.values[3];
	packet.flags = task.values[4] ? AV_PKT_FLAG_KEY : 0;
	packet.flags |= task.values[5] ? AV_PKT_FLAG_DISCARDABLE : 0;

	if (packet.pts == -1)
		packet.pts = AV_NOPTS_VALUE;
	if (packet.dts == -1)
		packet.dts = packet.dts;

	if (packet.pts != AV_NOPTS_VALUE)
		packet.pts = av_rescale_q(packet.pts, st->codec->time_base, st->time_base);
	if (packet.dts != AV_NOPTS_VALUE)
		packet.dts = av_rescale_q(packet.dts, st->codec->time_base, st->time_base);
	if (packet.duration > 0 && (packet.duration & 0x40000000) == 0)
		packet.duration = av_rescale_q(packet.duration, st->codec->time_base, st->time_base);

	marker_duration = (packet.duration & 0x40000000) != 0;
	packet.duration &= ~0x40000000;

	result.ret = av_write_frame(context_format, &packet);

	av_free_packet(&packet);
}

// =======================================================================================================================================================================================================================================================
// STATIC METHOD - DRONE - COMMAND
// =======================================================================================================================================================================================================================================================

static void drone_format();

static void drone_command(int *reinit)
{
	drone_pipe_read(&task, sizeof(struct COMMAND_TASK));

	*reinit = task.input == ULLONG_MAX && task.output == ULLONG_MAX;

	if (*reinit)
	{
		drone_reset();
		return;
	}

	drone_transfer_input();

	if (init.type == TYPE_CONTAINER)
		drone_format();
	else
		drone_codec();

	drone_transfer_output();
}

// =======================================================================================================================================================================================================================================================
// STATIC METHOD - DRONE - FORMAT
// =======================================================================================================================================================================================================================================================

static void drone_format()
{
	if (task.output_size == ULONG_MAX - 3)
		drone_format_stream();
	else if (task.output_size == ULONG_MAX - 2)
		drone_format_header();
	else if (task.output_size == ULONG_MAX - 1)
		drone_format_footer();
	else
		drone_format_packet();
}

// =======================================================================================================================================================================================================================================================
// STATIC METHOD - DRONE - INIT
// =======================================================================================================================================================================================================================================================

static void drone_init()
{
	enum AVCodecID requested_codec = 0;
	const char* requested_format = NULL;

	drone_pipe_read(&init, sizeof(struct COMMAND_INIT));

	if (init.signature != 'chmd' || init.version != 20150507) // chmd => Chameleon Drone
		av_quit(1);

	switch (init.codec)
	{
		case 'h265':   requested_codec = AV_CODEC_ID_HEVC;         break;
		case 'h264':   requested_codec = AV_CODEC_ID_H264;         break;
		case 'h262':   requested_codec = AV_CODEC_ID_MPEG2VIDEO;   break;
		case 'vc1':    requested_codec = AV_CODEC_ID_VC1;          break;

		case 'hevc':   requested_codec = AV_CODEC_ID_HEVC;         break;
		case 'avc':    requested_codec = AV_CODEC_ID_H264;         break;

		case 'mp2':    requested_codec = AV_CODEC_ID_MP2;          break;
		case 'mp3':    requested_codec = AV_CODEC_ID_MP3;          break;
		case 'aac':    requested_codec = AV_CODEC_ID_AAC;          break;
		case 'dts':    requested_codec = AV_CODEC_ID_DTS;          break;
		case 'ac3':    requested_codec = AV_CODEC_ID_AC3;          break;
		case 'eac3':   requested_codec = AV_CODEC_ID_EAC3;         break;
		case 'mlp':    requested_codec = AV_CODEC_ID_TRUEHD;       break;

		case 'dvd':    requested_codec = AV_CODEC_ID_DVD_SUBTITLE; break;
		case 'pg':     requested_codec = AV_CODEC_ID_HDMV_PGS_SUBTITLE; break;

			// containers
		case 'avi':    requested_format = "avi";                   break;
		case 'asf':    requested_format = "asf";                   break;
		case 'mov':    requested_format = "mov";                   break;
		case 'mp4':  requested_format = "mp4";                   break;
		case 'psp':  requested_format = "psp";                   break;
		case 'ipod': requested_format = "ipod";                  break;
		case 'mkv':    requested_format = "matroska";              break;
		case 'ts':     requested_format = "mpegts";                break;
		case 'tv':     requested_format = "mpegts";                break;
	}

	if (init.type == TYPE_CONTAINER)
	{
		format = av_guess_format(requested_format, NULL, NULL);

		if (!format)
			av_quit(2);

		context_format = avformat_alloc_context();
		context_format->oformat = format; // the avformat API is idiotic

		if (strcmp(requested_format, "mpegts") == 0)
		{
			av_opt_set(context_format->priv_data, "mpegts_m2ts_mode", "1", 0);
		}
	}
	else
	{
		codec = init.type == TYPE_DECODER ? avcodec_find_decoder(requested_codec) : avcodec_find_encoder(requested_codec);
		context_codec = avcodec_alloc_context3(codec);

		//memset(VIDEO_SURFACE + 1920 * 1080 * 0, 0xCC, 1920 * 1080);
		//memset(VIDEO_SURFACE + 1920 * 1080 * 1, 0xCC, 1920 * 1080);

		// codec not found
		if (!codec)
		{
			AVCodec *c;

			for (c = av_codec_next(NULL); c; c = av_codec_next(c))
			{
				av_log(NULL, AV_LOG_ERROR, "codec: %s (%s) ; %u, %u\n", c->name, c->long_name, c->type, c->id);
			}

			for (c = av_codec_next(NULL); c; c = av_codec_next(c))
			{
				av_log(NULL, AV_LOG_ERROR, "codec: %s (%s) ; %u, %u\n", c->name, c->long_name, c->type, c->id);
			}

			av_quit(2);
		}

		// parameters
		if (init.values[7] & 0x1000)
		{
			drone_pipe_read(&paramlen, 2);

			params = (char *) malloc(paramlen + 4);
			drone_pipe_read(params, paramlen);

			*(int *) (params + paramlen) = 0;
		}
	}
}

// =======================================================================================================================================================================================================================================================
// STATIC METHOD - DRONE - STARTUP
// =======================================================================================================================================================================================================================================================

static void drone_startup()
{
	stdout->_file = stderr->_file;
	av_log_set_level(AV_LOG_INFO);

	CreateThread(NULL, 0, &ThreadAlive, NULL, 0, NULL);

	av_register_all(); // only way to fly, implicitly registers codecs
}

// =======================================================================================================================================================================================================================================================
// METHOD - MAIN
// =======================================================================================================================================================================================================================================================

int main()
{
	int reinit = 1;

	DebugActivateWER();
	drone_startup();

	for (; ; )
	{
		if (reinit)
			drone_init();

		drone_command(&reinit);
	}

	(void) drone_debugger_attach;
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// END OF MODULE
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~