/*
 * This file is part of Libav.
 *
 * Libav is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * Libav is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with Libav; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

#include "config.h"
#include <windows.h>
//#include <math.h>

#include "libavutil/mathematics.h"  // av_rescale_q
#include "libavutil/channel_layout.h"
#include "libavformat/avformat.h"
#include "libavcodec/avcodec.h"
#include "cmdutils.h"

int ff_find_stream_index(AVFormatContext *s, int id); // internal
AVChapter *avpriv_new_chapter(AVFormatContext *s, int id, AVRational time_base, int64_t start, int64_t end, const char *title); // internal

#pragma GCC diagnostic ignored "-Wmultichar"
#pragma GCC diagnostic ignored "-Wstrict-prototypes"
#pragma GCC diagnostic ignored "-Wint-to-pointer-cast"
#pragma GCC diagnostic ignored "-Wdeprecated-declarations"

const char program_name[] = "ffdrone";
const int program_birth_year = 2012;

void show_help_default(const char *opt, const char *arg)
{
}

// =======================================================================================================================================================================================================================================================
// STRUCTURES
// =======================================================================================================================================================================================================================================================

struct COMMAND_INIT
{
	int32_t signature;
	int32_t version;
	int32_t share_input;
	int32_t share_output;

	int32_t type;
	int32_t codec;
	int32_t threads;
	int32_t reserved;

	int32_t values[32];
};

struct COMMAND_TASK
{
	int64_t input;
	int64_t output;

	int32_t input_len;
	int32_t output_size;

	int32_t values[32];
};

struct RESULT_TASK
{
	int32_t ret;
	int32_t output_len;

	int32_t values[32];
};

#define TYPE_CONTAINER  0
#define TYPE_DECODER    1
#define TYPE_ENCODER    2

// =======================================================================================================================================================================================================================================================
// GLOABALS
// =======================================================================================================================================================================================================================================================

struct COMMAND_INIT init;
struct COMMAND_TASK task;
struct RESULT_TASK result;

static char *params;
static short paramlen;

static AVCodecContext *context_codec;
static AVFormatContext *context_format;
static int context_format_busy;

static AVCodec *codec;
static AVOutputFormat *format;

static AVPacket packet;
static AVFrame frame;

#define av_quit(code)        ExitProcess(0xDEAD0000 + code)

// =======================================================================================================================================================================================================================================================
// STATIC METHOD - DRONE - RESET
// =======================================================================================================================================================================================================================================================

static void drone_reset() 
{
	if (init.type == TYPE_CONTAINER)
	{
		avformat_free_context(context_format);

		context_format = NULL, format = NULL, context_format_busy = 0;
	}
	else
	{
		avcodec_close(context_codec);
		av_free(context_codec);
		free(params);

		context_codec = NULL, codec = NULL, params = NULL;
	}
}

// =======================================================================================================================================================================================================================================================
// STATIC METHOD - DRONE - PIPE - READ
// =======================================================================================================================================================================================================================================================

static void drone_pipe_read(void *data, int length)
{
	static HANDLE Pipe; int have, pos; 

	if (!Pipe)
		Pipe = GetStdHandle(STD_INPUT_HANDLE);

	for (pos = 0 ; pos < length ; ++pos)
	{
		if (!ReadFile(Pipe, (byte *) data + pos, length - pos, (DWORD *) &have, NULL) || have == 0)
			av_quit(11);

		pos += have;
	}
}

// =======================================================================================================================================================================================================================================================
// STATIC METHOD - DRONE - PIPE - WRITE
// =======================================================================================================================================================================================================================================================

static void drone_pipe_write(const void *data, int length)
{
	static HANDLE Pipe; int have; 

	if (!Pipe)
		Pipe = GetStdHandle(STD_OUTPUT_HANDLE);

	if (!WriteFile(Pipe, data, length, (DWORD *) &have, NULL))
		av_quit(12);
}

// =======================================================================================================================================================================================================================================================
// STATIC METHOD - DRONE - TRANSFER - INPUT
// =======================================================================================================================================================================================================================================================

static void drone_transfer_input() 
{
	if (!init.share_input && task.input_len <= LONG_MAX) // using pipe
	{
		static byte *memory_input; 

		if (!memory_input) 
			memory_input = av_malloc(1024 << 10);

		task.input = (int64_t) memory_input;

		drone_pipe_read(memory_input, task.input_len);
	}

	if (!init.share_output)
	{
		static byte *memory_output;

		if (!memory_output) 
			memory_output = av_malloc(1024 << 10);

		task.output = (int64_t) memory_output;
	}
}

// =======================================================================================================================================================================================================================================================
// STATIC METHOD - DRONE - TRANSFER - OUTPUT
// =======================================================================================================================================================================================================================================================

static void drone_transfer_output()
{
	byte *output = (byte *) task.output;

	// transfer into shared memory (video)
	if (init.share_output && init.type == TYPE_DECODER)
	{
		int p, plane = 0, plane_sec = 0;

		if (codec->type == AVMEDIA_TYPE_VIDEO && frame.linesize[0])
			plane = frame.linesize[0] * frame.height, plane_sec = plane / 4;
		if (codec->type == AVMEDIA_TYPE_AUDIO && frame.linesize[0])
			plane = frame.linesize[0], plane_sec = plane;

		for (p = 0 ; p < 8 ; ++p, plane = plane_sec)
		{
			if (frame.data[p]) memcpy(output, frame.data[p], plane), output += plane;
		}
	}

	drone_pipe_write(&result, sizeof(struct RESULT_TASK));
}

// =======================================================================================================================================================================================================================================================
// STATIC METHOD - DRONE - CODEC - INIT - DECODE - VIDEO
// =======================================================================================================================================================================================================================================================

static void drone_codec_decode_video() 
{
	int got = 0, pitch, plane;

	packet.data = task.input_len ? (byte *) task.input : NULL; // special case for decoder flush
	packet.size = task.input_len;

	result.ret = avcodec_decode_video2(context_codec, &frame, &got, &packet);
	av_log(NULL, AV_LOG_ERROR, "Ret = %d, Got = %d\n", result.ret, got);

	if (got && result.ret != packet.size && (result.ret & 0x40000000) == 0)
		result.ret = AVERROR_INVALIDDATA;

	pitch = got ? frame.linesize[0] : 0;
	plane = pitch * context_codec->height / 4;

	result.output_len = plane * 6; // YUV 420 

	result.values[0] = pitch;
	result.values[1] = frame.key_frame;
	result.values[2] = frame.coded_picture_number;
	result.values[3] = frame.display_picture_number;
	result.values[4] = context_codec->frame_number;
}

// =======================================================================================================================================================================================================================================================
// STATIC METHOD - DRONE - CODEC - DECODE - AUDIO
// =======================================================================================================================================================================================================================================================

static void drone_codec_decode_audio() 
{
	int got = 0, pitch;

	packet.data = task.input_len ? (byte *) task.input : NULL; // special case for decoder flush
	packet.size = task.input_len;

	result.ret = avcodec_decode_audio4(context_codec, &frame, &got, &packet);

	if (got && result.ret != packet.size)
		result.ret = AVERROR_INVALIDDATA;

	pitch = got ? frame.linesize[0] : 0;

	result.output_len = pitch;
}

// =======================================================================================================================================================================================================================================================
// STATIC METHOD - DRONE - CODEC - ENCODE - VIDEO
// =======================================================================================================================================================================================================================================================

static void drone_codec_encode_video() 
{
	int got = 0, pitch = task.values[1], plane; byte *input;

	plane = pitch * context_codec->height / 4;
	input = (byte *) task.input;

	frame.data[0] = input;
	frame.data[1] = input + plane * 4;
	frame.data[2] = input + plane * 5;

	frame.linesize[0] = pitch;
	frame.linesize[1] = pitch / 2;
	frame.linesize[2] = pitch / 2;

	frame.pts = task.values[0]; // libx264 requires pts

	packet.data = (byte *) task.output;
	packet.size = task.output_size;

	result.ret = avcodec_encode_video2(context_codec, &packet, task.input_len ? &frame : NULL, &got);
	result.output_len = packet.size;

	result.values[0] = (int32_t) (size_t) context_codec->extradata;
	result.values[1] = context_codec->extradata_size;
	result.values[2] = context_codec->coded_frame &&  context_codec->coded_frame->pict_type;
	result.values[3] = context_codec->coded_frame && (packet.flags & AV_PKT_FLAG_KEY) != 0;
	result.values[4] = packet.pts;
	result.values[5] = packet.dts;
}

// =======================================================================================================================================================================================================================================================
// STATIC METHOD - DRONE - CODEC - ENCODE - SUBTITLE
// =======================================================================================================================================================================================================================================================

static void drone_codec_encode_subtitle() // time, duration, left, top, width, height,  
{
	AVSubtitle sub;
	AVSubtitleRect rect;
	AVSubtitleRect* rects[1];

	sub.format = 0;
	sub.start_display_time = 0;
	sub.end_display_time = task.values[1];
	sub.num_rects = 1;
	sub.rects = rects;
	sub.pts = task.values[0];

	rects[0] = &rect;

	rect.x = task.values[2];
	rect.y = task.values[3];
	rect.w = task.values[4];
	rect.h = task.values[5];
	rect.nb_colors = 4;
	rect.type = SUBTITLE_BITMAP;
	rect.text = NULL;
	rect.ass = NULL;
	rect.flags = 0;

	rect.pict.data[0] = (uint8_t *) task.input;
	rect.pict.data[1] = (uint8_t *) task.input + rect.w * rect.h;
	rect.pict.linesize[0] = rect.w;
	rect.pict.linesize[1] = 4;

	result.ret = avcodec_encode_subtitle(context_codec, task.output, task.output_size, &sub);

	if (result.ret > 0)
		result.output_len = result.ret, result.ret = 0;
	else
		result.output_len = 0;
}

// =======================================================================================================================================================================================================================================================
// STATIC METHOD - DRONE - CODEC - ENCODE - AUDIO
// =======================================================================================================================================================================================================================================================

static void drone_codec_encode_audio() 
{
	int got = 0, samples = task.values[1], plane; byte *input;

	plane = samples * sizeof(float);
	input = (byte *) task.input;

	for (int nn = 0 ; nn < 8 ; ++nn)
	{
		frame.data[nn]     = input + plane * nn;
		frame.linesize[nn] = plane;
	}

	frame.nb_samples = samples;

	packet.data = (byte *) task.output;
	packet.size = task.output_size;

	result.ret = avcodec_encode_audio2(context_codec, &packet, &frame, &got);
	result.output_len = packet.size;

	// manually set
	packet.pts = context_codec->frame_number;
	packet.dts = 0xFFFFFFFF;

	result.values[0] = (int32_t)(size_t) context_codec->extradata;
	result.values[1] = context_codec->extradata_size;
	result.values[2] = context_codec->coded_frame && context_codec->coded_frame->pict_type;
	result.values[3] = context_codec->coded_frame && context_codec->coded_frame->key_frame;
	result.values[4] = packet.pts;
	result.values[5] = packet.dts;
}

// =======================================================================================================================================================================================================================================================
// STATIC METHOD - DRONE - CODEC - DECODE - VIDEO - INIT
// =======================================================================================================================================================================================================================================================

static void drone_codec_decode_video_init(AVDictionary **dict) // VALUES: width, height
{
	// decoder VC-1 ==> a real pain in the ass
	if (init.codec == AV_CODEC_ID_VC1)
	{
		context_codec->width = init.values[0];
		context_codec->height = init.values[1];
		context_codec->extradata = (byte *) task.input;
		context_codec->extradata_size = task.input_len;
	}

	// decoder H.264 ==> needs some speedup
	if (init.codec == AV_CODEC_ID_H264)
	{
		context_codec->flags2 |= CODEC_FLAG2_FAST; // Cannot parallelize deblocking type 1, decoding such frames in sequential order -> switching
	}

	if (init.values[2] & 1)
	{
		//context_codec->flags |= CODEC_FLAG_OUTPUT_DECODER_ORDER;
	}
}

// =======================================================================================================================================================================================================================================================
// STATIC METHOD - DRONE - CODEC - DECODE - AUDIO - INIT
// =======================================================================================================================================================================================================================================================

static void drone_codec_decode_audio_init(AVDictionary **dict) // VALUES: channels
{
	context_codec->request_channels = init.values[0];
	context_codec->request_channel_layout = 
		context_codec->request_channels == 6 ? AV_CH_LAYOUT_5POINT1 : 
		context_codec->request_channels == 2 ? AV_CH_LAYOUT_STEREO : 
		context_codec->request_channels == 1 ? AV_CH_LAYOUT_MONO : 0;
}

// =======================================================================================================================================================================================================================================================
// STATIC METHOD - DRONE - CODEC - ENCODE - VIDEO - INIT
// =======================================================================================================================================================================================================================================================

static void drone_codec_encode_video_init(AVDictionary **dict) // VALUES: width, height, aspect-x, aspect-y, rate-num, rate-den, bitrate
{
	context_codec->width = init.values[0];
	context_codec->height = init.values[1];
	context_codec->pix_fmt = PIX_FMT_YUV420P;
	context_codec->sample_aspect_ratio.num = init.values[2];
	context_codec->sample_aspect_ratio.den = init.values[3];
	context_codec->time_base.num = init.values[5]; // time-base, not fps
	context_codec->time_base.den = init.values[4];
	context_codec->bit_rate = init.values[6];
	context_codec->flags |= CODEC_FLAG_QSCALE;

	context_codec->coder_type = FF_CODER_TYPE_AC;
	context_codec->gop_size = 24;
	context_codec->max_b_frames = 2; 

	if (init.values[7] & 1)
		context_codec->flags |= CODEC_FLAG_GLOBAL_HEADER;

	context_codec->sample_aspect_ratio = av_mul_q(context_codec->sample_aspect_ratio, (AVRational){ context_codec->height, context_codec->width }); // revert sar->dar

	// DICTIONARY
  // ==========

	if (init.values[7] & 0x1000)
	{
		char *parameters = params; int len = paramlen, pos; 

		for (pos = 0 ; pos < len ; )
		{
			int a, b, flags;

			a = strlen(parameters + pos);
			if (pos + a + 1 > len)
				break;

			b = strlen(parameters + pos + a + 1);
			if (pos + a + 1 + b + 1 + 4 > len)
				break;

			flags = *(int *)(parameters + pos + a + 1 + b + 1);

			av_dict_set(dict, parameters + pos, parameters + pos + a + 1, flags);

			pos += a + 1 + b + 1 + 4;
		}
	}	
}

// =======================================================================================================================================================================================================================================================
// STATIC METHOD - DRONE - CODEC - ENCODE - AUDIO - INIT
// =======================================================================================================================================================================================================================================================

static void drone_codec_encode_audio_init(AVDictionary **dict) // // VALUES: channels, sample-rate, bitrate
{
	context_codec->sample_fmt = AV_SAMPLE_FMT_FLTP;
	context_codec->channels = init.values[0];
	context_codec->channel_layout = context_codec->channels == 6 ? AV_CH_LAYOUT_5POINT1 : context_codec->channels == 2 ? AV_CH_LAYOUT_STEREO : context_codec->channels == 1 ? AV_CH_LAYOUT_MONO : 0;
	context_codec->sample_rate = init.values[1];
	context_codec->bit_rate = init.values[2];

	if (init.values[3] & 1)
		context_codec->flags |= CODEC_FLAG_GLOBAL_HEADER;

	context_codec->strict_std_compliance = FF_COMPLIANCE_EXPERIMENTAL; // we are heroes
}

// =======================================================================================================================================================================================================================================================
// STATIC METHOD - DRONE - CODEC - ENCODE - SUBTITLE - INIT
// =======================================================================================================================================================================================================================================================

static void drone_codec_encode_subtitle_init() // VALUES: none
{
}

// =======================================================================================================================================================================================================================================================
// STATIC METHOD - DRONE - CODEC - INIT
// =======================================================================================================================================================================================================================================================

static void drone_codec_init() 
{
	AVDictionary *dict = NULL;

	if (context_codec->codec)
		return;

	context_codec->debug |= FF_DEBUG_PICT_INFO;
	context_codec->thread_count = init.threads;

	if (init.type == TYPE_DECODER && codec->type == AVMEDIA_TYPE_VIDEO)
		drone_codec_decode_video_init(&dict);
	if (init.type == TYPE_DECODER && codec->type == AVMEDIA_TYPE_AUDIO)
		drone_codec_decode_audio_init(&dict);

	if (init.type == TYPE_ENCODER && codec->type == AVMEDIA_TYPE_VIDEO)
		drone_codec_encode_video_init(&dict);
	if (init.type == TYPE_ENCODER && codec->type == AVMEDIA_TYPE_AUDIO)
		drone_codec_encode_audio_init(&dict);
	if (init.type == TYPE_ENCODER && codec->type == AVMEDIA_TYPE_SUBTITLE)
		drone_codec_encode_subtitle_init(&dict);

	result.ret = avcodec_open2(context_codec, codec, &dict);

	if (result.ret != 0 || !context_codec->codec)
		av_quit(3);

	av_dict_free(&dict);
}

// =======================================================================================================================================================================================================================================================
// STATIC METHOD - DRONE - CODEC
// =======================================================================================================================================================================================================================================================

static void drone_codec() 
{
	drone_codec_init();

	// data is cloned, we must not free it
	frame.nb_side_data = 0;
	frame.side_data = NULL; 
	av_frame_unref(&frame);

	av_init_packet(&packet);
	avcodec_get_frame_defaults(&frame);

	if (init.type == TYPE_DECODER && codec->type == AVMEDIA_TYPE_VIDEO)
		drone_codec_decode_video();
	if (init.type == TYPE_DECODER && codec->type == AVMEDIA_TYPE_AUDIO)
		drone_codec_decode_audio();

	if (init.type == TYPE_ENCODER && codec->type == AVMEDIA_TYPE_VIDEO)
		drone_codec_encode_video();
	if (init.type == TYPE_ENCODER && codec->type == AVMEDIA_TYPE_AUDIO)
		drone_codec_encode_audio();
	if (init.type == TYPE_ENCODER && codec->type == AVMEDIA_TYPE_SUBTITLE)
		drone_codec_encode_subtitle();
}

// =======================================================================================================================================================================================================================================================
// STATIC METHOD - DRONE - FORMAT - STREAM
// =======================================================================================================================================================================================================================================================

static void drone_format_stream() 
{
	AVStream *st; AVCodec *codec; int id, codec_id;

	id = task.values[0];
	codec_id = task.values[1];

	switch (codec_id)
	{
		case 'h262': codec_id = AV_CODEC_ID_MPEG2VIDEO;  break;
		case 'h264': codec_id = AV_CODEC_ID_H264;        break;
		case 'vc1':  codec_id = AV_CODEC_ID_VC1;         break;

		case 'mp2':  codec_id = AV_CODEC_ID_MP2;         break;
		case 'mp3':  codec_id = AV_CODEC_ID_MP3;         break;
		case 'aac':  codec_id = AV_CODEC_ID_AAC;         break;
		case 'ac3':  codec_id = AV_CODEC_ID_AC3;         break;
		case 'dts':  codec_id = AV_CODEC_ID_DTS;         break;

		case 'dvd':  codec_id = AV_CODEC_ID_DVD_SUBTITLE; break;
	}

	// id, codec_id, frame_rate_n, frame_rate_d
	codec = avcodec_find_encoder(codec_id);
	if (!codec)
		codec = avcodec_find_decoder(codec_id);
	if (!codec)
		av_quit(4);

	if (context_format_busy)
		av_quit(0x103);

	st = avformat_new_stream(context_format, codec);
	st->id = id;
	st->codec->codec_type = codec->type;
	st->codec->codec_id = codec->id;
	st->codec->time_base.num = task.values[3]; // time-base, not fps
	st->codec->time_base.den = task.values[2];

	if (!st->codec->time_base.num)
		av_quit(0x101);
	if (!st->codec->time_base.den)
		av_quit(0x102);

	if (task.input)
	{
		st->codec->flags |= CODEC_FLAG_GLOBAL_HEADER;
		st->codec->extradata_size = task.input_len;
		st->codec->extradata = av_malloc(task.input_len);

		memcpy(st->codec->extradata, (byte *) task.input, task.input_len);
	}

	// width, height, aspect_ratio_x, aspect_ratio_y
	if (codec->type == AVMEDIA_TYPE_VIDEO)
	{
		st->codec->width = task.values[4];
		st->codec->height = task.values[5];
		st->codec->sample_aspect_ratio.num = task.values[6];
		st->codec->sample_aspect_ratio.den = task.values[7];

		if (!st->codec->sample_aspect_ratio.num)
			av_quit(0x105);
		if (!st->codec->sample_aspect_ratio.den)
			av_quit(0x106);

		// from DAR to SAR
		st->codec->sample_aspect_ratio = av_mul_q(st->codec->sample_aspect_ratio, (AVRational){ st->codec->height, st->codec->width });

		st->sample_aspect_ratio = st->codec->sample_aspect_ratio; // have to make avformat_write_header() happy
	}

	// sampling_rate, channels, frame_size
	if (codec->type == AVMEDIA_TYPE_AUDIO)
	{
		st->codec->bits_per_raw_sample = 16;
		st->codec->sample_rate = task.values[4];
		st->codec->channels = task.values[5];
		st->codec->frame_size = task.values[6];

		if (!st->codec->frame_size)
			av_quit(0x111);
		if (!st->codec->sample_rate)
			av_quit(0x112);
		if (!st->codec->channels)
			av_quit(0x113);
	}

	// parameters
	if (task.values[8] & 0x1000)
	{
		char *parameters; int len = 0, pos; 

		// read
		drone_pipe_read(&len, 2);

		parameters = (char *) malloc(len + 4);
		drone_pipe_read(parameters, len);

		*(int *)(parameters + len) = 0;

		// extract
		for (pos = 0 ; pos < len ; )
		{
			int a, b, flags;

			a = strlen(parameters + pos);
			if (pos + a + 1 > len)
				break;

			b = strlen(parameters + pos + a + 1);
			if (pos + a + 1 + b + 1 + 4 > len)
				break;

			flags = *(int *)(parameters + pos + a + 1 + b + 1);

			av_dict_set(&st->metadata, parameters + pos, parameters + pos + a + 1, flags);

			pos += a + 1 + b + 1 + 4;
		}

		free(parameters);
	}	

	result.ret = 0;
}

// =======================================================================================================================================================================================================================================================
// STATIC METHOD - DRONE - FORMAT - HEADER
// =======================================================================================================================================================================================================================================================

static void drone_format_header() 
{
	char filename[128];
	AVDictionary *options = NULL;

	// parameters
	if (task.values[1] & 0x1000)
	{
		char *parameters; int len = 0, pos; 

		// read
		drone_pipe_read(&len, 2);

		parameters = (char *) malloc(len + 4);
		drone_pipe_read(parameters, len);

		*(int *)(parameters + len) = 0;

		// extract
		for (pos = 0 ; pos < len ; )
		{
			int a, b, flags;

			a = strlen(parameters + pos);
			if (pos + a + 1 > len)
				break;

			b = strlen(parameters + pos + a + 1);
			if (pos + a + 1 + b + 1 + 4 > len)
				break;

			flags = *(int *)(parameters + pos + a + 1 + b + 1);

			if (strcmp(parameters + pos, "chapter") == 0)
			{
				int index = flags;
				int64_t start = *(int64_t *)(parameters + pos + a + 1 + b + 1 + 4);
				int64_t end = *(int64_t *)(parameters + pos + a + 1 + b + 1 + 4 + 8);
				const char *title = parameters + pos + a + 1;

				avpriv_new_chapter(context_format, index, (AVRational) { 1, 10000000 }, start, end, title);

				pos += 8 + 8;
			}
			else
			{
				av_dict_set(&options, parameters + pos, parameters + pos + a + 1, flags);
			}

			pos += a + 1 + b + 1 + 4;
		}

		free(parameters);
	}	

	sprintf(filename, "\\device\\handle\\%u", task.values[0]);

	result.ret = avio_open(&context_format->pb, filename, AVIO_FLAG_WRITE);
	if (result.ret != 0)
		return;

	result.ret = avformat_write_header(context_format, &options);
	
	av_dict_free(&options);
}

// =======================================================================================================================================================================================================================================================
// STATIC METHOD - DRONE - FORMAT - FOOTER
// =======================================================================================================================================================================================================================================================

static void drone_format_footer() 
{
	result.ret = av_write_trailer(context_format);
	if (result.ret != 0)
		return;

	avio_close(context_format->pb);
}

// =======================================================================================================================================================================================================================================================
// STATIC METHOD - DRONE - FORMAT - PACKET
// =======================================================================================================================================================================================================================================================

static void drone_format_packet() 
{
	AVStream *st; int id, idx, size; byte *data;

	context_format_busy = 1;

	idx = ff_find_stream_index(context_format, task.values[0]);
	if (idx == -1)
		av_quit(0x121);

	id = task.values[0];
	st = context_format->streams[idx];

	if (!st)
		av_quit(0x122);

	size = task.input_len;
	data = av_malloc(size);
	if (!data)
		av_quit(0x1A0);

	memcpy(data, (byte *) task.input, size);

	av_init_packet(&packet);
	packet.buf = av_buffer_create(data, size, av_buffer_default_free, NULL, 0);
	packet.data = packet.buf->data;
	packet.size = packet.buf->size;

	packet.stream_index = st->index;
	packet.pts = task.values[1];
	packet.dts = task.values[2];
	packet.duration = task.values[3];
	packet.flags = task.values[4] ? AV_PKT_FLAG_KEY : 0;

	if (packet.pts == -1)
		packet.pts = AV_NOPTS_VALUE;
	if (packet.dts == -1)
		packet.dts = packet.dts;

	if (packet.pts != AV_NOPTS_VALUE)
		packet.pts = av_rescale_q(packet.pts, st->codec->time_base, st->time_base);
	if (packet.dts != AV_NOPTS_VALUE)
		packet.dts = av_rescale_q(packet.dts, st->codec->time_base, st->time_base);
	if (packet.duration > 0)
		packet.duration = av_rescale_q(packet.duration, st->codec->time_base, st->time_base);

	result.ret = av_write_frame(context_format, &packet);

	av_free_packet(&packet);
}

// =======================================================================================================================================================================================================================================================
// STATIC METHOD - DRONE - FORMAT
// =======================================================================================================================================================================================================================================================

static void drone_format() 
{
	if (task.output_size == ULONG_MAX - 3)
		drone_format_stream();
	else if (task.output_size == ULONG_MAX - 2)
		drone_format_header();
	else if (task.output_size == ULONG_MAX - 1)
		drone_format_footer();
	else
		drone_format_packet();
}

// =======================================================================================================================================================================================================================================================
// STATIC METHOD - DRONE - COMMAND
// =======================================================================================================================================================================================================================================================

static void drone_command(int *reinit) 
{
	drone_pipe_read(&task, sizeof(struct COMMAND_TASK));

	*reinit = task.input == ULLONG_MAX && task.output == ULLONG_MAX;

	if (*reinit)
	{
		drone_reset();
		return;
	}

	drone_transfer_input();

	if (init.type == TYPE_CONTAINER)
		drone_format();
	else
		drone_codec();

	drone_transfer_output();
}

// =======================================================================================================================================================================================================================================================
// STATIC METHOD - DRONE - INIT
// =======================================================================================================================================================================================================================================================

static void drone_init() 
{
	drone_pipe_read(&init, sizeof(struct COMMAND_INIT));

	if (init.signature != 'chmd' || init.version != 20150507) // chmd => Chameleon Drone
		av_quit(1);

	switch (init.codec)
	{
		case 'h264': init.codec = AV_CODEC_ID_H264;        break;
		case 'h262': init.codec = AV_CODEC_ID_MPEG2VIDEO;  break;
		case 'vc1':  init.codec = AV_CODEC_ID_VC1;         break;

		case 'mp2':  init.codec = AV_CODEC_ID_MP2;         break;
		case 'mp3':  init.codec = AV_CODEC_ID_MP3;         break;
		case 'aac':  init.codec = AV_CODEC_ID_AAC;         break;
		case 'dts':  init.codec = AV_CODEC_ID_DTS;         break;
		case 'ac3':  init.codec = AV_CODEC_ID_AC3;         break;
		case 'eac3': init.codec = AV_CODEC_ID_EAC3;        break;
		case 'mlp':  init.codec = AV_CODEC_ID_TRUEHD;      break;

		case 'dvd':  init.codec = AV_CODEC_ID_DVD_SUBTITLE; break;

		// containers
		case 'avi':  init.codec = (int32_t) "avi";        break;
		case 'asf':  init.codec = (int32_t) "asf";        break;
		case 'mov':  init.codec = (int32_t) "mov";        break;
			case 'mp4':  init.codec = (int32_t) "mp4";      break;
			case 'psp':  init.codec = (int32_t) "psp";      break;
			case 'ipod': init.codec = (int32_t) "ipod";     break;
		case 'mkv':  init.codec = (int32_t) "matroska";   break;
		case 'ts':   init.codec = (int32_t) "mpegts";     break;
	}

	if (init.type == TYPE_CONTAINER)
	{
		format = av_guess_format((const char *) init.codec, NULL, NULL);

		if (!format)
			av_quit(2);

		context_format = avformat_alloc_context();
		context_format->oformat = format; // the avformat API is idiotic
	}
	else
	{
		codec = init.type == TYPE_DECODER ? avcodec_find_decoder(init.codec) : avcodec_find_encoder(init.codec);
		context_codec = avcodec_alloc_context3(NULL); // avcodec_open() called later, need to initialize parameters

		if (!codec)
		{
			AVCodec *c;
			
			for (c = av_codec_next(NULL) ; c ; c = av_codec_next(c))
			{
				av_log(NULL, AV_LOG_ERROR, "codec: %s (%s) ; %u, %u\n", c->name, c->long_name, c->type, c->id);
			}

			av_quit(2);
		}

		// parameters
		if (init.values[7] & 0x1000)
		{
			drone_pipe_read(&paramlen, 2);

			params = (char *) malloc(paramlen + 4);
			drone_pipe_read(params, paramlen);

			*(int *)(params + paramlen) = 0;
		}	
	}
}

// =======================================================================================================================================================================================================================================================
// STATIC METHOD - DRONE - STARTUP
// =======================================================================================================================================================================================================================================================

static void drone_startup() 
{
	stdout->_file = stderr->_file;
	av_log_set_level(AV_LOG_DEBUG);

	av_register_all(); // only way to fly, implicitly registers codecs
}

// =======================================================================================================================================================================================================================================================
// METHOD - MAIN
// =======================================================================================================================================================================================================================================================

int main()
{
	int reinit = 1;

	drone_startup();

	for ( ; ; )
	{
		if (reinit)
			drone_init();
		
		drone_command(&reinit);
	}
}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// END OF MODULE
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~